package extension

import (
	"reflect"
	"strings"
	"testing"

	"git.perx.ru/perxis/perxis-go/pkg/errors"
	"git.perx.ru/perxis/perxis-go/pkg/extension/mocks"
	"github.com/stretchr/testify/mock"
)

func TestGetResults(t *testing.T) {

	getMockExtension := func(name string, wantErr ...bool) Extension {
		ext := &mocks.Extension{}
		ext.On("GetDescriptor").Return(&ExtensionDescriptor{
			Extension:   name,
			Title:       strings.ToTitle(name),
			Description: "test extension",
			Version:     "0.0.0",
		})

		var err error
		if len(wantErr) > 0 {
			err = errors.WithDetail(errors.New("some err"), "Ошибка")
		}

		ext.On("Install", mock.Anything, mock.Anything).Return(err)
		return ext
	}

	tests := []struct {
		name       string
		services   []Extension
		extensions []string
		fn         func(svc Extension) error
		want       []*RequestResult
	}{
		{
			name:       "one extension without errors",
			services:   []Extension{getMockExtension("a"), getMockExtension("b")},
			extensions: []string{"a"},
			fn:         func(svc Extension) error { return nil },
			want: []*RequestResult{
				{Extension: "a", State: RequestOK},
			},
		},
		{
			name:       "multiple extensions without errors",
			services:   []Extension{getMockExtension("a"), getMockExtension("b"), getMockExtension("c")},
			extensions: []string{"a", "c"},
			fn:         func(svc Extension) error { return nil },
			want: []*RequestResult{
				{Extension: "a", State: RequestOK},
				{Extension: "c", State: RequestOK},
			},
		},
		{
			name:       "multiple extensions, one returns error",
			services:   []Extension{getMockExtension("a"), getMockExtension("b"), getMockExtension("c", true)},
			extensions: []string{"a", "c"},
			fn:         func(svc Extension) error { return svc.Install(nil, nil) },
			want: []*RequestResult{
				{Extension: "a", State: RequestOK},
				{Extension: "c", State: RequestError, Error: "some err", Msg: "Ошибка\n"},
			},
		},
		{
			name:       "multiple extensions, all return error",
			services:   []Extension{getMockExtension("a", true), getMockExtension("b", true), getMockExtension("c", true)},
			extensions: []string{"a", "b", "c"},
			fn:         func(svc Extension) error { return svc.Install(nil, nil) },
			want: []*RequestResult{
				{Extension: "a", State: RequestError, Error: "some err", Msg: "Ошибка\n"},
				{Extension: "b", State: RequestError, Error: "some err", Msg: "Ошибка\n"},
				{Extension: "c", State: RequestError, Error: "some err", Msg: "Ошибка\n"},
			},
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			srv := NewServer(tt.services...)
			if got := srv.getResults(tt.extensions, tt.fn); !reflect.DeepEqual(got, tt.want) {
				t.Errorf("getResults() = %v, want %v", got, tt.want)
			}
		})
	}
}
