package setup

import (
	"context"
	"strings"

	"git.perx.ru/perxis/perxis-go/pkg/errors"
	"git.perx.ru/perxis/perxis-go/pkg/items"
	"go.uber.org/zap"
)

var (
	ErrCheckItems     = errors.New("items check error")
	ErrInstallItems   = errors.New("failed to install items")
	ErrUninstallItems = errors.New("failed to uninstall items")
	ErrItemsNotFound  = errors.New("item not found")
)

type ItemsOption func(c *ItemConfig)
type UpdateItemFn func(s *Setup, exist, new *items.Item) (*items.Item, bool)
type DeleteItemFn func(s *Setup, col *items.Item) bool

type ItemConfig struct {
	item     *items.Item
	UpdateFn UpdateItemFn
	DeleteFn DeleteItemFn
}

func NewItemConfig(item *items.Item, opt ...ItemsOption) ItemConfig {
	c := ItemConfig{item: item}

	KeepExistingItem()(&c)
	DeleteItemIfRemove()(&c)

	for _, o := range opt {
		o(&c)
	}

	return c
}

func OverwriteItem() ItemsOption {
	return func(c *ItemConfig) {
		c.UpdateFn = func(s *Setup, old, new *items.Item) (*items.Item, bool) { return new, true }
	}
}

func KeepExistingItem() ItemsOption {
	return func(c *ItemConfig) {
		c.UpdateFn = func(s *Setup, old, new *items.Item) (*items.Item, bool) { return old, false }
	}
}

func DeleteItem() ItemsOption {
	return func(c *ItemConfig) {
		c.DeleteFn = func(s *Setup, item *items.Item) bool { return true }
	}
}

func DeleteItemIfRemove() ItemsOption {
	return func(c *ItemConfig) {
		c.DeleteFn = func(s *Setup, item *items.Item) bool { return s.IsRemove() }
	}
}

func (s *Setup) InstallItems(ctx context.Context) error {
	if len(s.Items) == 0 {
		return nil
	}

	s.logger.Debug("Install items", zap.Int("Items", len(s.Items)))

	var errs []error
	for _, c := range s.Items {
		if err := s.InstallItem(ctx, c); err != nil {
			errs = append(errs, errors.WithDetailf(err, "Возникла ошибка при добавлении элемента %s(%s)", c.item.ID, c.item.CollectionID))
			s.logger.Error("Failed to install item",
				zap.String("Item", c.item.ID),
				zap.String("Item", c.item.CollectionID),
				zap.Error(err),
			)
		}
	}

	if len(errs) > 0 {
		return errors.WithErrors(ErrInstallItems, errs...)
	}

	return nil
}

func (s *Setup) InstallItem(ctx context.Context, c ItemConfig) error {
	item := c.item
	item.SpaceID, item.EnvID = s.SpaceID, s.EnvironmentID

	exist, err := s.content.Items.Get(ctx, item.SpaceID, item.EnvID, item.CollectionID, item.ID)
	if err != nil && !strings.Contains(err.Error(), items.ErrNotFound.Error()) {
		return err
	}
	if exist == nil {
		return s.CreateAndPublishItem(ctx, item)
	}

	if item, changed := c.UpdateFn(s, exist, c.item); changed {
		return s.UpdateAndPublishItem(ctx, item)
	}

	return nil
}

func (s *Setup) UninstallItems(ctx context.Context) error {
	if len(s.Items) == 0 {
		return nil
	}

	s.logger.Debug("Uninstall items", zap.Int("Items", len(s.Items)))

	var errs []error
	for _, c := range s.Items {
		if err := s.UninstallItem(ctx, c); err != nil {
			errs = append(errs, errors.WithDetailf(err, "Возникла ошибка при удалении элемента %s(%s)", c.item.ID, c.item.CollectionID))
			s.logger.Error("Failed to uninstall item",
				zap.String("Item", c.item.ID),
				zap.String("Item", c.item.CollectionID),
				zap.Error(err),
			)
		}
	}

	if len(errs) > 0 {
		return errors.WithErrors(ErrUninstallItems, errs...)
	}

	return nil
}

func (s *Setup) UninstallItem(ctx context.Context, c ItemConfig) error {
	if c.DeleteFn(s, c.item) {
		err := s.content.Items.Delete(ctx, s.SpaceID, s.EnvironmentID, c.item.CollectionID, c.item.ID)
		if err != nil && !strings.Contains(err.Error(), items.ErrNotFound.Error()) {
			return err
		}
	}
	return nil
}

func (s *Setup) CheckItems(ctx context.Context) error {
	if len(s.Items) == 0 {
		return nil
	}

	var errs []error
	s.logger.Debug("Check items", zap.Int("Items", len(s.Items)))

	for _, c := range s.Items {
		if err := s.CheckItem(ctx, c); err != nil {
			errs = append(errs, errors.WithDetailf(err, "Не найден элемент %s(%s)", c.item.ID, c.item.CollectionID))
		}
	}

	if len(errs) > 0 {
		return errors.WithErrors(ErrCheckItems, errs...)
	}

	return nil
}

func (s *Setup) CheckItem(ctx context.Context, c ItemConfig) (err error) {
	_, err = s.content.Items.Get(ctx, s.SpaceID, s.EnvironmentID, c.item.CollectionID, c.item.ID)
	return err
}

func (s *Setup) CreateAndPublishItem(ctx context.Context, item *items.Item) error {
	var err error
	if item, err = s.content.Items.Create(ctx, item); err != nil {
		return errors.Wrap(err, "create item")
	}
	if err = s.content.Items.Publish(ctx, item); err != nil {
		return errors.Wrap(err, "publish item")
	}
	return nil
}

func (s *Setup) UpdateAndPublishItem(ctx context.Context, item *items.Item) error {
	var err error
	if err = s.content.Items.Update(ctx, item); err != nil {
		return errors.Wrap(err, "update item")
	}
	if err = s.content.Items.Publish(ctx, item); err != nil {
		return errors.Wrap(err, "publish item")
	}
	return nil
}

func (s *Setup) removeItems(collID string) {
	itms := make([]ItemConfig, 0, len(s.Items))
	for _, i := range s.Items {
		if i.item.CollectionID != collID {
			itms = append(itms, i)
		}
	}
	s.Items = itms
}
